/*******************************************************************\

Module: 

Author: CM Wintersteiger

\*******************************************************************/


// old, slow function dependency check functions

/*******************************************************************\

Function: QBV::check_function_dependencies

  Inputs: model

 Outputs: none

 Purpose: Checks for cyclic dependencies in model functions

\*******************************************************************/

void QBV::check_function_dependencies(std::map<unsigned, signed> &model) const
{
  if(debug)
  {
    std::cout << "Checking function dependencies ..." << std::endl;
    qcnf.show_quantification_sequence(std::cout);
  }
  
  std::map<unsigned, dep_set_type > dependencies;
  std::vector<bool> last_changes;
  last_changes.resize(qcnf.max_variable_no+1, false);
  
  // ITE extension dependencies
  for(model_cache::const_iterator it=model_cache_ite.begin();
      it!=model_cache_ite.end();
      it++)
  {
    const std::vector<Literal> &m=it->second;
    
    assert(m.size()==3);
                        
    dependencies[it->first].insert(m[0].variable());
    dependencies[it->first].insert(dependencies[m[0].variable()].begin(),
                                   dependencies[m[0].variable()].end());
    
    dependencies[it->first].insert(m[1].variable());
    dependencies[it->first].insert(dependencies[m[1].variable()].begin(),
                                   dependencies[m[1].variable()].end());
    
    dependencies[it->first].insert(m[2].variable());
    dependencies[it->first].insert(dependencies[m[2].variable()].begin(),
                                   dependencies[m[2].variable()].end());
    
    last_changes[it->first]=true;
  }
  
  // AND extension dependencies
  for(model_cache::const_iterator it=model_cache_and.begin();
      it!=model_cache_and.end();
      it++)
  {
    const std::vector<Literal>& m=it->second;          
              
    for(std::vector<Literal>::const_iterator ait=m.begin();
        ait!=m.end();
        ait++)
    {
      unsigned v=ait->variable();
      dependencies[it->first].insert(v);
      dependencies[it->first].insert(dependencies[v].begin(),
                                     dependencies[v].end());
      
      last_changes[it->first]=true;
    }
  }
  
  dependency_fixpoint(dependencies, last_changes);
  
  for(std::vector<QCNF::qscopet>::const_iterator it=qcnf.quantification_sequence.begin();
      it!=qcnf.quantification_sequence.end();
      it++)
  {
    if(it->size()==0) continue;
    if(qcnf.is_universal(it->front().variable)) continue;
        
    // Model dependencies
    for(std::list<Quantifier>::const_iterator qit=it->begin();
        qit!=it->end();
        qit++)
    {      
      unsigned v=qit->variable;
      if(model.find(v)!=model.end())
      {
        dependencies[v].insert(Literal(model[v]).variable());        
        last_changes[v]=true;
      }
    }
    
    dependency_fixpoint(dependencies, last_changes);
  }
  
  find_cyclic_dependencies(dependencies);
  
  #if false
  std::cout << "Dependencies: " << std::endl;
  for(std::map<unsigned, dep_set_type >::const_iterator it=dependencies.begin();
      it!=dependencies.end();
      it++)
  {
    std::cout << it->first << ": ";
    for(dep_set_type::const_iterator dit=it->second.begin();
        dit!=it->second.end();
        dit++)
    {
      if(dit!=it->second.begin()) std::cout << ",";
      std::cout << *dit;
    }
    std::cout << std::endl;
  }
  #endif
}

/*******************************************************************\

Function: QBV::dependency_fixpoint

  Inputs: 

 Outputs: 

 Purpose: 

\*******************************************************************/

void QBV::dependency_fixpoint(
  std::map<unsigned, dep_set_type > &dependencies,
  std::vector<bool> &last_changes) const
{   
  // std::cout << "Fixpoint..." << std::endl;
  
  bool progress=true;
  while(progress)
  {
    //std::cout << "."; std::cout.flush();
    std::vector<unsigned> difference;
    std::vector<bool> current_changes;
    current_changes.resize(last_changes.size(), false);
    
    for(std::map<unsigned, dep_set_type >::const_iterator it=dependencies.begin();
        it!=dependencies.end();
        it++)
    {
      unsigned cur=it->first;
      dep_set_type &curset=dependencies[cur];      

      for(dep_set_type::const_iterator sit=curset.begin();
          sit!=curset.end();
          sit++)
      {  
        if(last_changes[*sit])
        {
          dep_set_type &dset=dependencies[*sit];
#if true
          difference.resize(dset.size(), 0);
          
          std::vector<unsigned>::iterator end_it=
            set_difference(dset.begin(), dset.end(),
                           curset.begin(), curset.end(),                       
                           difference.begin());
          
          if((end_it - difference.begin())>0)
          {
            curset.insert(difference.begin(), end_it);
            current_changes[cur]=true;
          }
#else
          bool changed=false;
          for(dep_set_type::const_iterator sit=dset.begin();
              sit!=dset.end();
              sit++)
          {
            if(curset.find(*sit)==curset.end())
            {
              std::pair<dep_set_type::const_iterator, bool> res=
                curset.insert(*sit);
              
              if(res.second)
                changed=true;
            }
          }
          
          if(changed)
            current_changes[cur]=true;
#endif
        }
      }      
    }
    
    progress=false;
    for(std::vector<bool>::const_iterator it=current_changes.begin();
        it!=current_changes.end();
        it++)
      if(*it) { progress=true; break; }
          
    last_changes.swap(current_changes);
    current_changes.clear();
  }
}

/*******************************************************************\

Function: QBV::find_cyclic_dependencies

  Inputs: 

 Outputs: 

 Purpose: 

\*******************************************************************/

void QBV::find_cyclic_dependencies(std::map<unsigned, dep_set_type > &dependencies) const
{
  for(std::map<unsigned, dep_set_type >::const_iterator it=dependencies.begin();
      it!=dependencies.end();
      it++)
  {
    for(dep_set_type::const_iterator sit=it->second.begin();
        sit!=it->second.end();
        sit++)
    {
      if(*sit==it->first)
      {
        std::stringstream s;
        s << "The model function for " << it->first << " is cyclic;" <<
          " it depends on the following variables: " << std::endl;
        
        for(dep_set_type::const_iterator dit=it->second.begin();
            dit!=it->second.end();
            dit++)
        {
          if(dit!=it->second.begin()) s << ",";
          s << *dit;
        }
        
        throw Exception(s.str());
      }
      else if(!qcnf.is_defined(*sit))
      {
        std::stringstream s;
        s << "The model function for " << it->first << 
        " depends on undefined variable " << *sit << std::endl;        
        
        throw Exception(s.str());
      }
    }
  }
}


#if false // an older version

/*******************************************************************\

Function: QBV::push_stack

  Inputs: stack, variable

 Outputs: none

 Purpose: Checks if a stack contains a variable, and if not pushes it 
          on the stack (for cyclic function definition checks)

\*******************************************************************/

bool QBV::push_stack(unsigned cur, std::vector<std::pair<unsigned, bool> > &stack, unsigned v) const
{
  assert(v<=qcnf.max_variable_no);
  
  if(qcnf.is_universal(v)) return false;
  
  std::vector<std::pair<unsigned, bool> >::const_iterator fit=stack.end();
  for(std::vector<std::pair<unsigned, bool> >::const_iterator sit=stack.begin();
      sit!=stack.end();
      sit++)
    if(sit->second && sit->first==v) { fit=sit; break; }
  
  if(fit!=stack.end())
  {
    std::stringstream s;
    bool first=true;
    
    //std::cout << "FIT: " << fit->first << " " << fit->second << std::endl;
    
    s << "The model function for " << cur << " is cyclic: " << std::endl;
    for(std::vector<std::pair<unsigned, bool> >::const_iterator sit=stack.begin();
        sit!=stack.end();
        sit++)
    {      
        if(!first) s << " -> "; else first=false;      
        s << sit->first;
    }
    
    s << " -> " << v;
    
    throw Exception(s.str());
  }
  
  stack.push_back(std::make_pair(v, false));
  return true;
}

/*******************************************************************\

Function: QBV::check_function_dependencies

  Inputs: model

 Outputs: none

 Purpose: Checks for cyclic dependencies in model functions

\*******************************************************************/

void QBV::check_function_dependencies(std::map<unsigned, signed> &model) const
{
  if(debug)
  {
    std::cout << "Checking function dependencies ..." << std::endl;
    qcnf.show_quantification_sequence(std::cout);
  }
  
  for(std::map<unsigned, signed>::const_iterator it=model.begin();
      it!=model.end();
      it++)
  {
    std::vector<std::pair<unsigned, bool> > stack;    
    stack.push_back(std::make_pair(it->first, false));
        
    while(!stack.empty())
    {
      Literal cur = stack.back().first;
      stack.back().second=true;
      unsigned var = cur.variable();      

      #if false
      std::cout << "Stack: ";      
      for(std::vector<std::pair<unsigned, bool> >::const_iterator pit=stack.begin();
          pit!=stack.end();
          pit++)
      {
        if(pit!=stack.begin()) std::cout << " -> ";
        std::cout << pit->first << ((pit->second)?"*":"");
      }
      std::cout << std::endl;
      #endif
      
      #if false
      std::cout << "Stack: ";  
      std::cout << stack.front().first << ":";
      for(std::vector<std::pair<unsigned, bool> >::const_iterator pit=stack.begin();
          pit!=stack.end();
          pit++)
      {        
        std::cout << ((pit->second)?"*":"-");
      }
      std::cout << std::endl;
      #endif
      
      if(qcnf.is_universal(var))
      { /* Nothing */ }
      else
      {
        std::map<unsigned, signed>::const_iterator model_it=model.find(var);
        model_cache::const_iterator ite_it=model_cache_ite.find(var);
        model_cache::const_iterator and_it=model_cache_and.find(var);
        
        if(model_it!=model.end())
          push_stack(it->first, stack, Literal(model_it->second).variable());                    
        else if(ite_it!=model_cache_ite.end())
        {
          const std::vector<Literal> &m=ite_it->second;
          std::vector<Literal>::const_iterator mit=m.begin();
          
          assert(m.size()==3);
                              
          push_stack(it->first, stack, m[0].variable());
          
          if(m[0].variable()!=m[1].variable())
            push_stack(it->first, stack, m[1].variable());
          
          if(m[0].variable()!=m[2].variable() &&
             m[1].variable()!=m[2].variable())
            push_stack(it->first, stack, m[2].variable());
        }
        else if(and_it!=model_cache_and.end())
        {
          const std::vector<Literal>& m=and_it->second;          
          
          for(std::vector<Literal>::const_iterator ait=m.begin();
              ait!=m.end();
              ait++)
            push_stack(it->first, stack, ait->variable());
        }        
      }
      
      while(!stack.empty() && stack.back().second)
        stack.pop_back();
    }
  }
}
#endif
