/*******************************************************************\

Module: QBV, The Quantified Boolean Validator, Main Module

Author: CM Wintersteiger

\*******************************************************************/

#include <stdlib.h>
#include <string.h>
#include <iostream>

#include "qbv.h"
#include "exception.h"
#include "version.h"

void usage( void )
{ 
  std::cout << "Usage: qbv [options] [qbf-filename] <certificate-filename>" << std::endl;
  std::cout << 
  "available options:" << std::endl <<
  "   --h, -h                show help" << std::endl <<
  "   --d                    turn on debug info" << std::endl <<  
  "   --ignore               ignore resolutions without antecedents" << std::endl <<
  "   --v                    show version information" << std::endl;
  std::cout.flush();
  exit(0);
}

void show_version( void )
{ 
  std::cout << QBV_VERSION << " (compiled " << QBV_COMPILATION_DATE << 
    ")" << std::endl;
  std::cout.flush();
  exit(0);
}

/*******************************************************************\
 
Function: main
 
  Inputs: argc/argv
 
 Outputs: true for a correctly validated certificate, false otherwise
 
 Purpose: Starts a validation.
 
\*******************************************************************/

int main(int argc, const char **argv)
{ 
  QBV qbv;
  
  bool result=false;
  std::string qbffn, qbcfn;

  for (int i=1; i<argc; i++)
  {
    if (strcmp(argv[i], "--h")==0 ||
        strcmp(argv[i], "-h")==0)
    {
      usage(); // exits!
    }
    else if (strcmp(argv[i], "--v")==0)
    {
      show_version(); //exits!
    }
    else if (strcmp(argv[i], "--d")==0)
    {        
      qbv.set_debug(true);
    }
    else if (strcmp(argv[i], "--ignore")==0)
    {
      qbv.set_ignore_no_reason();
    }
    else if (qbffn=="")      
      qbffn = argv[i];
    else if (qbcfn=="")
      qbcfn = argv[i];
    else
      usage(); // exits!
  }
  
  if (qbffn.empty() && qbcfn.empty())
  {
    std::cout << "Error: No certificate filename given." << std::endl;
    usage(); // exits!
  }
  
  try {
    if (qbcfn.empty())
      result = qbv.validate("", qbffn);
    else
      result = qbv.validate(qbffn, qbcfn);    
  }
  catch (Exception &e)
  {
    std::cout << "Error at line " << qbv.cur_linenumber << ", column " <<
      qbv.cur_colnumber << ": " << e.message() << std::endl;
    return 30;
  }
  
  if (result)
  {
    std::cout << "Certificate validated correctly." << std::endl;
    return 10;
  }
  else
  {
    std::cout << "Certificate erroneous." << std::endl;
    return 20;
  }
}
